import argparse
import dataclasses
import json
import signal
import threading
from datetime import datetime, timedelta, timezone
from typing import Optional, List

import requests
import time

from heal_check.exceptions import CheckWarning, UrlWarning, ResponseWarning, TimestampWarning, StatusWarning, ExpectedModeWarning, DictionaryError, StatusError, WarningTimeoutError

REQUEST_TIMEOUT_DEFAULT: float = 3.05  # see https://docs.python-requests.org/en/master/user/advanced/#timeouts
HEAL_TIMEOUT_DEFAULT: float = 15
WARNING_TIMEOUT_DEFAULT: float = 600  # 10 minutes
DELAY_DEFAULT: float = 3


@dataclasses.dataclass
class RemoteData:
    timestamp: datetime
    status: str
    mode: str

    def to_json(self):
        return json.dumps(dataclasses.asdict(self), indent=2, default=lambda obj: obj.isoformat())


def _response_from_url(url: str, request_timeout: float = REQUEST_TIMEOUT_DEFAULT) -> str:
    try:
        response = requests.get(url, timeout=request_timeout)
        response.raise_for_status()
        return response.text
    except Exception as exception:
        raise UrlWarning(url, request_timeout) from exception


def _dictionary_from_response(response: str) -> dict:
    try:
        return json.loads(response)
    except Exception as exception:
        raise ResponseWarning(response) from exception


def _remote_data_from_dictionary(dictionary: dict) -> RemoteData:
    try:
        return RemoteData(datetime.fromisoformat(dictionary["timestamp"]), str(dictionary["status"]), str(dictionary["mode"]))
    except Exception as exception:
        raise DictionaryError(dictionary) from exception


def _check_timestamp(timestamp: datetime, heal_timeout: float = HEAL_TIMEOUT_DEFAULT, now: datetime = None) -> None:
    if not now:
        now = datetime.fromtimestamp(time.time(), timezone.utc)

    if now - timestamp > timedelta(seconds=heal_timeout):
        raise TimestampWarning(timestamp, heal_timeout, now)


def _check_status(status: str) -> None:
    if status == "fixing":
        raise StatusWarning()
    elif status != "ok":
        raise StatusError(status)


def check(url: str, request_timeout: float = REQUEST_TIMEOUT_DEFAULT, heal_timeout: float = HEAL_TIMEOUT_DEFAULT) -> RemoteData:
    response = _response_from_url(url, request_timeout)
    dictionary = _dictionary_from_response(response)
    remote_data = _remote_data_from_dictionary(dictionary)
    _check_status(remote_data.status)
    _check_timestamp(remote_data.timestamp, heal_timeout)

    return remote_data


def warning_tolerant_check(url: str,
                           event: threading.Event = None,
                           expected_mode: str = None,
                           delay: float = DELAY_DEFAULT,
                           warning_timeout: float = WARNING_TIMEOUT_DEFAULT,
                           request_timeout: float = REQUEST_TIMEOUT_DEFAULT,
                           heal_timeout: float = HEAL_TIMEOUT_DEFAULT) -> RemoteData:
    if not event:
        event = threading.Event()

    timer = threading.Timer(warning_timeout, lambda: event.set())
    timer.start()

    try:
        while True:
            try:
                remote_data = check(url, request_timeout, heal_timeout)
                if not expected_mode or remote_data.mode == expected_mode:
                    return remote_data
                else:
                    raise ExpectedModeWarning(expected_mode)
            except CheckWarning as warning:
                event.wait(delay)
                if event.is_set():
                    raise WarningTimeoutError(warning)
    finally:
        timer.cancel()


def main(args: Optional[List[str]] = None) -> None:
    argparser = argparse.ArgumentParser(description="")
    argparser.add_argument("url", help="", metavar="<url>")
    argparser.add_argument("-m", "--expected-mode", help="", metavar="<mode>")
    argparser.add_argument("-d", "--delay", type=float, default=DELAY_DEFAULT, help="", metavar="<duration>")
    argparser.add_argument("-w", "--warning-timeout", type=float, default=WARNING_TIMEOUT_DEFAULT, help="", metavar="<duration>")
    argparser.add_argument("-r", "--request-timeout", type=float, default=REQUEST_TIMEOUT_DEFAULT, help="", metavar="<duration>")
    argparser.add_argument("-t", "--heal-timeout", type=float, default=HEAL_TIMEOUT_DEFAULT, help="", metavar="<duration>")
    args = argparser.parse_args(args)

    # turning SIGINT and SIGTERM into an event
    event = threading.Event()
    for sig in [signal.SIGINT, signal.SIGTERM]:
        signal.signal(sig, lambda signum, frame: event.set())

    print(warning_tolerant_check(args.url, event, args.expected_mode, args.delay, args.warning_timeout, args.request_timeout, args.heal_timeout).to_json())
